# Import libraries
from collections.abc import AsyncIterator
from websockets.asyncio.client import connect, ClientConnection
from websockets.protocol import State
import asyncio
import json

# Import I4.0 utilities
from .Config import Conf as Config
from .Service import Service as Service
from .Service import ServiceManager as ServiceManager
from . import Encryption

# Servers variables
ServersTasks: dict[int, list[Service]] = {}
Connected: tuple[str, str, int] | None = None
Conf: Config = Config()
ClientVersion: int = 150000

# WebSocket variables
ClientSocket: ClientConnection | None = None

# Keys
PublicKey: bytes | None = None
PrivateKey: bytes | None = None
ServerPublicKey: bytes | None = None

def GetTasks() -> dict[int, list[Service]]:
    # Return a copy of the tasks
    return ServersTasks.copy()

async def __connect_str__(Server: str, Port: int = 8060) -> None:
    global ClientSocket, Connected, PublicKey, PrivateKey, ServerPublicKey

    # Check if the keys have been generated
    if (PublicKey is None or PrivateKey is None):
        # Generate keys
        PrivateKey, PublicKey = Encryption.CreateKeys()

    # Disconnect from any other server
    await Disconnect()

    # Get the extension
    if (Server.startswith("ws://")):
        ext = "ws://"
        Server = Server[5:]
    elif (Server.startswith("wss://")):
        ext = "wss://"
        Server = Server[6:]
    else:
        ext = "ws://"
    
    # Get port
    if (Server.count(":") > 0):
        Server = Server.split(":")
        Port = int(Server[1])
        Server = Server[0]

    # Create socket and connect
    ClientSocket = await connect(f"{ext}{Server}:{Port}", max_size = None, ping_interval = None, ping_timeout = None, close_timeout = None)

    # Wait until it's connected
    waited = 0

    while (ClientSocket.state == State.CONNECTING):
        if (waited > 50):
            raise TimeoutError("Server not responding.")

        await asyncio.sleep(0.1)
        waited += 1
    
    # Check client state
    if (ClientSocket.state != State.OPEN):
        raise RuntimeError(f"Error connecting. WebSocket state: {ClientSocket.state}")

    # Set connected server
    Connected = (Server, f"{ext}{Server}:{Port}", -1)

    # Get the public key from the server
    if (ServerPublicKey is None):
        ServerPublicKey = await SendAndReceive("get_public_key".encode("utf-8"), False)
    
    # Get the server version from the server
    serverVersion = await SendAndReceive("get_server_version".encode("utf-8"), False)
    serverVersion = serverVersion.decode("utf-8")

    try:
        Connected = (Connected[0], Connected[1], int(serverVersion))
    except:
        pass

async def __connect_int__(Server: int, Port: int = 8060) -> None:
    if (Server < 0 or Server >= len(Conf.Servers)):
        # The server ID is invalid
        raise Exception("Invalid server ID.")

    # Connect to the server
    await __connect_str__(Conf.Servers[Server], Port)

async def Connect(Server: str | int, Port: int = 8060) -> None:
    """
    This will connect the user to a server.
    If the user is alredy connected to a server, this will disconnect it first.
    """

    if (type(Server) == str):
        # Connect to a server by string
        await __connect_str__(Server, Port)
    elif (type(Server) == int):
        # Connect to a server by ID
        await __connect_int__(Server, Port)
    else:
        # Invalid server type
        raise Exception("Invalid server type.")

async def Disconnect() -> None:
    """
    This will disconnect the user from the server.
    If the user is not connected to any server this will not do anything.
    """
    global ClientSocket, Connected, ServerPublicKey

    if (ClientSocket != None):
        try:
            # Try to close the connection to the server
            await ClientSocket.close()
        except:
            # Ignore any errors during the disconnection
            pass

        # Delete the socket
        ClientSocket = None

    # Delete the connected server
    Connected = None

    # Delete the server public key
    ServerPublicKey = None

async def __receive_from_server__() -> bytes:
    # Receive from the server
    data = await ClientSocket.recv()

    # Check the data type
    if (type(data) == str):
        # It's a string, replace with encoded data
        data = data.encode("utf-8")
    elif (type(data) != bytes):
        # Invalid type
        raise Exception("Invalid received type.")

    # Return the bytes
    return data

async def SendAndReceive(Data: bytes | str, Encrypt: bool = True) -> bytes:
    """
    This will send a message to the server and wait for it's response.
    The user must be connected to a server before using this.
    """

    # Check if you're connected
    if (not IsConnected()):
        await Disconnect()
        raise Exception("Connect to a server first.")
    
    # Convert the data to bytes
    if (isinstance(Data, str)):
        Data = Data.encode("utf-8")

    # Encrypt the data
    if (Encrypt):
        Data = Encryption.EncryptMessage(Data, ServerPublicKey, Encryption.__parse_hash__(Conf.HashAlgorithm), Connected[2] >= 140100)
        Data = json.dumps({"Hash": Conf.HashAlgorithm, "Message": Data, "Version": str(ClientVersion)})
    else:
        Data = Data.decode("utf-8") if (isinstance(Data, bytes)) else Data

    # Send the data
    await ClientSocket.send(Data)

    received = await __receive_from_server__()

    # Decrypt the data
    if (Encrypt):
        received = Encryption.DecryptMessage(received, PrivateKey, Encryption.__parse_hash__(Conf.HashAlgorithm)).encode("utf-8")

    # Return the response
    return received

async def GetServicesFromServer() -> list[Service]:
    """
    This will get all the available services from the current connected server.
    The user must be connected to a server before using this.
    """
    # Ask the server for the models, then deserialize the response into a dictionary and create a services list
    received = await SendAndReceive(json.dumps({
        "Service": "get_all_services",
        "Prompt": "",
        "Files": {},
        "PublicKey": PublicKey.decode("utf-8")
    }).encode("utf-8"), True)
    receivedData = received.decode("utf-8")
    jsonData = json.loads(receivedData)
    services = []

    # Set jsonData
    jsonData = json.loads(jsonData["response"])

    for service in list(jsonData.keys()):
        # Ignore if it alredy exists on the list
        if (ServiceManager.FromString(service) not in services):
            # Add the service to the list if don't exists
            services.append(ServiceManager.FromString(service))

    # Return all the services
    return services

async def FindFirstServer(ServiceToExecute: Service, DeleteData: bool = False) -> int:
    """
    This will find the first server available with the specified service.
    If there isn't any available server, this will return an error.
    """

    # Check if the user wants to delete the current services list or if the list is empty
    if (DeleteData or len(ServersTasks) == 0):
        # Clear the list and set an invalid server ID
        ServersTasks.clear()
        serverToReturn = -1

        for server in Conf.Servers:
            # For every server
            try:
                # Try to connect to the server
                await Connect(server)

                # Try to get all the services of the server, then add them to the list
                services = await GetServicesFromServer()
                ServersTasks[Conf.Servers.index(server)] = services

                # Check every service available from this server
                for service in services:
                    if (service == ServiceToExecute and serverToReturn < 0):
                        # If the service is the same the user requests, set the server ID to this server's ID
                        serverToReturn = Conf.Servers.index(server)

                # Disconnect from the server
                await Disconnect()
            except:
                # If the server doesn't responds or if there's another error, ignore it and continue with the next server
                ServersTasks[Conf.Servers.index(server)] = []

        # Check if the server ID is valid
        if (serverToReturn >= 0):
            # Return the ID if it's valid (the requested service has been found)
            return serverToReturn

        # Return an error if it's invalid (the requested service could not be found)
        raise Exception("Could not find any server.")

    # If the user doesn't want to delete the data and the service list isn't empty
    for server in Conf.Servers:
        # For every server get it's services
        for service in ServersTasks[Conf.Servers.index(server)]:
            # Check if the service is the same the user is requesting
            if (service == ServiceToExecute):
                # If it is, then return the server ID
                return Conf.Servers.index(server)

        # Continue with the next server if the requested service could not be found on this one

    # If the requested service has not been found on any server, return an error
    raise Exception("Could not find any server.")

def IsConnected() -> bool:
    # Define globals
    global Connected, ClientSocket

    # Check if the user is connected to a server
    return Connected is not None and ClientSocket != None and ClientSocket.state == State.OPEN

async def SendAndWaitForStreaming(Data: str | bytes) -> AsyncIterator[dict[str, any]]:
    """
    This will send a prompt to the connected server and will wait for the full response.
    Also, this will yield the response of the server when received even if they're not complete.
    """

    # Send to the server and get a response
    responseBytes = await SendAndReceive(Data.encode("utf-8"), True)
    responseStr = responseBytes.decode("utf-8")
    response = json.loads(responseStr)

    # Check if it's streaming
    while (not response["ended"]):
        # Hasn't ended stream, returning dictionary
        yield response

        # Wait for receive
        responseBytes = await __receive_from_server__()
        responseStr = Encryption.DecryptMessage(responseBytes, PrivateKey, Encryption.__parse_hash__(Conf.HashAlgorithm))
        response = json.loads(responseStr)

    yield response


class Conf():
    def __init__(self) -> None:
        # General
        self.Servers: list[str] = [
            "127.0.0.1",                # Localhost
            "main.tao71.org",           # TAO71 Main Server
            "tao71.sytes.net"           # TAO71 Main Server (old URL)
        ]
        self.ServerAPIKey: str = ""
        self.AllowedTools: list[str] | str | None = None
        self.ExtraTools: list[dict[str, str | dict[str, any]]] = []
        self.HashAlgorithm: str = "sha512"
        self.MaxLength: int | None = None
        self.Temperature: float | None = None
        self.SeeTools: bool = False
        self.AllowDataSave: bool = True
        self.DefaultIndex: int = -2

        # Chatbot
        self.Chatbot_ExtraSystemPrompts: str = ""
        self.Chatbot_AllowServerSystemPrompts: list[bool] | bool = [True, True, True, True]  # Boolean array for prompts [default I4.0 personality, server's extra system prompt (global), server's extra system prompt (for the selected index), server's dynamic system prompts]
        self.Chatbot_Conversation: str = "Client"
        self.Chatbot_AIArgs: str | None = None
        self.Chatbot_TopP: float | None = None
        self.Chatbot_TopK: int | None = None
        self.Chatbot_MinP: float | None = None
        self.Chatbot_TypicalP: float | None = None

        # Simulated vision v1
        self.SimulatedVision_v1_Image2Text_Allow: bool = True
        self.SimulatedVision_v1_Image2Text_Index: int | None = None
        self.SimulatedVision_v1_ObjectDetection_Allow: bool = True
        self.SimulatedVision_v1_ObjectDetection_Index: int | None = None

        # Simulated vision v2
        self.SimulatedVision_v2_Video_Allow: bool = True
        self.SimulatedVision_v2_Video_ProcessFrames: int = 1
        self.SimulatedVision_v2_Video_UseAudition: bool = True

        # Simulated audition v1
        self.SimulatedAudition_v1_SpeechToText_Allow: bool = True
        self.SimulatedAudition_v1_SpeechToText_Index: int | None = None

        # Text2Image
        self.Text2Image_Width: int = -1
        self.Text2Image_Height: int = -1
        self.Text2Image_GuidanceScale: float = -1
        self.Text2Image_Steps: int = -1

        # RVC
        self.RVC_FilterRadius: int = 3
        self.RVC_Protect: float = 0.33
        self.RVC_f0: int = 0
        self.RVC_IndexRate: float = 0.75
        self.RVC_MixRate: float = 0.25

        # UVR
        self.UVR_Agg: int = 10

        # Image2Image
        self.Image2Image_Steps: int = 10
    
    @classmethod
    def __from_dict__(cls, Dictionary: dict[str, any]) -> "Conf":
        obj = cls()
        
        for key, value in Dictionary.items():
            if (hasattr(obj, key)):
                setattr(obj, key, value)
        
        return obj
    
    def __to_dict__(self) -> dict[str, any]:
        return self.__dict__


# Import libraries
from threading import Event
import pyaudio
import numpy as np
import time
import wave
import io

def RAWToWAV(RawBytes: bytes, Channels: int = 1, SampleRate: int = 44100) -> bytes:
    """
    Converts raw audio bytes to WAV format.

    ## Parameters:
    - RawBytes (bytes): Raw audio bytes.

    ## Returns:
    bytes: WAV audio bytes.
    """

    # Create buffer
    buffer = io.BytesIO()

    # Write the WAV header
    with wave.open(buffer, "wb") as wavefile:
        wavefile.setnchannels(Channels)
        wavefile.setsampwidth(2)
        wavefile.setframerate(SampleRate)
        wavefile.writeframesraw(RawBytes)

    # Read the buffer bytes
    buffer.seek(0)
    bufBytes = buffer.read()

    # Close the buffer and return the bytes
    buffer.close()
    return bufBytes

def RecordMicrophoneUntilSilence(SilenceDuration: float = 1.5, VoiceThreshold: int = 150, Channels: int = 1, SampleRate: int = 44100) -> tuple[bool, bytes]:
    """
    Records the audio on your default microphone.

    ## Parameters:
    - SilenceDuration (float): Duration of silence in seconds to stop recording.
    - VoiceThreshold (int): Threshold for detecting silence.
    - Channels (int): Number of audio channels.
    - SampleRate (int): Sample rate of the audio.

    ## Returns:
    tuple: (containsVoice (bool), data (bytes))

    The bytes returned are RAW audio bytes.
    """

    # Create variables
    chunkSize = 1024
    format = pyaudio.paInt16
    frames = []
    silenceStart = None
    containsVoice = False
    stopEvent = Event()

    # Start PyAudeio
    p = pyaudio.PyAudio()

    # Start recording
    stream = p.open(
        format = format,
        channels = Channels,
        rate = SampleRate,
        input = True,
        frames_per_buffer = chunkSize
    )

    # While the event is not completed
    while (not stopEvent.is_set()):
        # Read the chunk data
        data = stream.read(chunkSize)

        # Save into the frames
        frames.append(data)

        # Convert the data to a numpy array for processing
        data = np.frombuffer(data, dtype = np.int16)
        silenceCount = np.sum(np.abs(data) < VoiceThreshold)

        # Check if the audio contains a voice
        if (silenceCount / len(data) >= 0.5):
            # It doesn't
            if (silenceStart is None):
                # Start the silence timer
                silenceStart = time.time()
            elif (time.time() - silenceStart > SilenceDuration):
                # Stop the recording
                stopEvent.set()
        else:
            # It does
            # Reset the timer and set the containsVoice bool to True
            silenceStart = None
            containsVoice = True
        
    # Stop the stream
    stream.stop_stream()
    stream.close()
    
    # Terminate PyAudio
    p.terminate()

    # Return the containsVoice boolean and the bytes of the audio
    return (containsVoice, b"".join(frames))


# Import libraries
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from concurrent.futures import ThreadPoolExecutor
import base64
import os
import json

def __parse_hash__(HashName: str) -> hashes.HashAlgorithm | None:
    if (HashName == "sha224"):
        return hashes.SHA224()
    elif (HashName == "sha256"):
        return hashes.SHA256()
    elif (HashName == "sha384"):
        return hashes.SHA384()
    elif (HashName == "sha512"):
        return hashes.SHA512()
    elif (HashName == "none"):
        return None
    
    raise ValueError("Unsupported hash algorithm.")

def __split_message__(Message: bytes, MaxSize: int) -> list[bytes]:
    return [Message[i:i + MaxSize] for i in range(0, len(Message), MaxSize)]

def CreateKeys() -> tuple[bytes, bytes]:
    prKey = rsa.generate_private_key(public_exponent = 65537, key_size = 2048)
    puKey = prKey.public_key()

    prKey = prKey.private_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PrivateFormat.PKCS8,
        encryption_algorithm = serialization.NoEncryption()
    )
    puKey = puKey.public_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PublicFormat.SubjectPublicKeyInfo
    )

    return (prKey, puKey)

def __is_hybrid_message__(Message: str) -> bool:
    try:
        parsed = json.loads(Message)
        return "key" in parsed and "data" in parsed and "iv" in parsed
    except:
        return False

def EncryptMessage(Message: str | bytes, Key: str | bytes, HashAlgorithm: hashes.HashAlgorithm | None, UseHybrid: bool = False) -> str:
    # Check if the hash algorithm is null
    if (HashAlgorithm is None):
        return Message.decode("utf-8") if (isinstance(Message, bytes)) else Message

    # Encrypt
    if (UseHybrid):
        return __encrypt_hybrid__(Message, Key, HashAlgorithm)
    else:
        return __encrypt_rsa__(Message, Key, HashAlgorithm)

def __encrypt_rsa__(Message: str | bytes, Key: str | bytes, HashAlgorithm: hashes.HashAlgorithm) -> str:
    if (isinstance(Message, str)):
        Message = Message.encode("utf-8")
    elif (not isinstance(Message, bytes)):
        raise Exception("Invalid message.")

    if (isinstance(Key, str)):
        Key = Key.encode("utf-8")
    elif (not isinstance(Key, bytes)):
        raise Exception(f"Invalid key. Key is '{type(Key)}'.")

    puKey = serialization.load_pem_public_key(Key)
    puKeySize = puKey.key_size // 8
    hashSize = HashAlgorithm.digest_size
    maxMessageSize = puKeySize - 2 * hashSize - 2
    
    chunks = __split_message__(Message, maxMessageSize)

    def enc(Chunk: bytes) -> str:
        encrypted = puKey.encrypt(
            Chunk,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = HashAlgorithm),
                algorithm = HashAlgorithm,
                label = None
            )
        )
        return base64.b64encode(encrypted).decode("utf-8")

    with ThreadPoolExecutor() as executor:
        encChunks = list(executor.map(enc, chunks))

    return "|".join(encChunks)

def __encrypt_hybrid__(Message: str | bytes, Key: str | bytes, HashAlgorithm: hashes.HashAlgorithm) -> str:
    if (isinstance(Message, str)):
        Message = Message.encode("utf-8")
    elif (not isinstance(Message, bytes)):
        raise Exception("Invalid message.")

    if (isinstance(Key, str)):
        Key = Key.encode("utf-8")
    elif (not isinstance(Key, bytes)):
        raise Exception("Invalid key.")

    puKey = serialization.load_pem_public_key(Key)
    aesKey = os.urandom(32)
    iv = os.urandom(16)

    # AES Encrypt
    cipher = Cipher(algorithms.AES(aesKey), modes.CBC(iv), backend = default_backend())
    encryptor = cipher.encryptor()
    padLength = 16 - (len(Message) % 16)
    paddedMessage = Message + bytes([padLength] * padLength)
    aesEncrypted = encryptor.update(paddedMessage) + encryptor.finalize()

    # RSA Encrypt key
    encryptedKey = puKey.encrypt(
        aesKey,
        padding.OAEP(
            mgf = padding.MGF1(algorithm = HashAlgorithm),
            algorithm = HashAlgorithm,
            label = None
        )
    )

    # Return as JSON base64
    return json.dumps({
        "key": base64.b64encode(encryptedKey).decode("utf-8"),
        "data": base64.b64encode(aesEncrypted).decode("utf-8"),
        "iv": base64.b64encode(iv).decode("utf-8")
    })

def DecryptMessage(Message: str | bytes, Key: str | bytes, HashAlgorithm: hashes.HashAlgorithm | None) -> str:
    if (isinstance(Message, bytes)):
        Message = Message.decode("utf-8")
    elif (not isinstance(Message, str)):
        raise RuntimeError("Invalid message.")

    if (isinstance(Key, str)):
        Key = Key.encode("utf-8")
    elif (not isinstance(Key, bytes)):
        raise RuntimeError("Invalid key.")

    if (HashAlgorithm is None):
        return Message

    if (__is_hybrid_message__(Message)):
        return __decrypt_hybrid__(Message, Key, HashAlgorithm)
    else:
        return __decrypt_rsa__(Message, Key, HashAlgorithm)

def __decrypt_rsa__(Message: str, Key: bytes, HashAlgorithm: hashes.HashAlgorithm) -> str:
    prKey = serialization.load_pem_private_key(Key, password=None)

    try:
        chunks = [base64.b64decode(chunk) for chunk in Message.split("|")] if ("|" in Message) else [base64.b64decode(Message)]
    except:
        return Message

    def dec(Chunk: bytes) -> bytes:
        return prKey.decrypt(
            Chunk,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = HashAlgorithm),
                algorithm = HashAlgorithm,
                label = None
            )
        )

    with ThreadPoolExecutor() as executor:
        decChunks = list(executor.map(dec, chunks))

    return b"".join(decChunks).decode("utf-8")

def __decrypt_hybrid__(Message: str, Key: bytes, HashAlgorithm: hashes.HashAlgorithm) -> str:
    prKey = serialization.load_pem_private_key(Key, password=None)
    parsed = json.loads(Message)

    aesKey = prKey.decrypt(
        base64.b64decode(parsed["key"]),
        padding.OAEP(
            mgf = padding.MGF1(algorithm = HashAlgorithm),
            algorithm = HashAlgorithm,
            label = None
        )
    )

    cipher = Cipher(
        algorithms.AES(aesKey),
        modes.CBC(base64.b64decode(parsed["iv"])),
        backend = default_backend()
    )

    decryptor = cipher.decryptor()
    padded = decryptor.update(base64.b64decode(parsed["data"])) + decryptor.finalize()
    padLength = padded[-1]

    return padded[:-padLength].decode("utf-8")


from enum import Enum

class Service(Enum):
    Chatbot = 0
    CustomCommand = 1
    ImageGeneration = 2
    ImageToText = 3
    SpeechToText = 4
    Audio = 5
    DepthEstimation = 6
    ObjectDetection = 7
    RVC = 8
    Translator = 9
    TextClassification = 10
    NSFWFilterText = 11
    NSFWFilterImage = 12
    UVR = 13
    ImageToImage = 14
    QuestionAnswering = 15
    LanguageDetection = 16
    NONE = -1

class ServiceManager:
    @staticmethod
    def FromString(ServiceName: str) -> Service:
        ServiceName = ServiceName.lower()
        mappings = {
            "chatbot": Service.Chatbot,
            "text2img": Service.ImageGeneration,
            "img2text": Service.ImageToText,
            "de": Service.DepthEstimation,
            "text2audio": Service.Audio,
            "speech2text": Service.SpeechToText,
            "od": Service.ObjectDetection,
            "rvc": Service.RVC,
            "tr": Service.Translator,
            "ld": Service.LanguageDetection,
            "sc": Service.TextClassification,
            "nsfw_filter-text": Service.NSFWFilterText,
            "nsfw_filter-image": Service.NSFWFilterImage,
            "uvr": Service.UVR,
            "img2img": Service.ImageToImage,
            "qa": Service.QuestionAnswering
        }

        if (ServiceName in mappings):
            return mappings[ServiceName]

        raise ValueError("Could not parse service.")

    @staticmethod
    def ToString(ServiceName: Service) -> str:
        mappings = {
            Service.Chatbot: "chatbot",
            Service.ImageGeneration: "text2img",
            Service.ImageToText: "img2text",
            Service.DepthEstimation: "de",
            Service.Audio: "text2audio",
            Service.SpeechToText: "speech2text",
            Service.ObjectDetection: "od",
            Service.RVC: "rvc",
            Service.Translator: "tr",
            Service.LanguageDetection: "ld",
            Service.TextClassification: "sc",
            Service.NSFWFilterText: "nsfw_filter-text",
            Service.NSFWFilterImage: "nsfw_filter-image",
            Service.UVR: "uvr",
            Service.ImageToImage: "img2img",
            Service.QuestionAnswering: "qa"
        }

        if (ServiceName in mappings):
            return mappings[ServiceName]

        raise ValueError("Could not parse service.")

    @staticmethod
    def FromInt(ServiceID: int) -> Service:
        try:
            return Service(ServiceID)
        except ValueError:
            raise ValueError(f"Could not find service with the ID '{ServiceID}'.")

    @staticmethod
    def ToInt(Serv: Service) -> int:
        return Serv.value

    @staticmethod
    def AutoConvert(Serv: str | int | Service) -> Service | str:
        if (type(Serv) == str):
            return ServiceManager.FromString(Serv)
        elif (type(Serv) == int):
            return ServiceManager.FromInt(Serv)
        elif (type(Serv) == Service):
            return ServiceManager.ToString(Serv)
        else:
            raise ValueError("Invalid service type.")



# Import libraries
from collections.abc import AsyncIterator
from io import BytesIO
from PIL import Image as PILImage
import av
import cv2
import json
import base64
import os

# Import I4.0 utilities
from .. import ServerConnection as ServerCon
from ..Config import Conf as Config
from ..Service import Service, ServiceManager

# Create variables
Conf: Config = Config()

class SimulatedVisionV1():
    @staticmethod
    def SeparateFiles(Files: list[dict[str, str]]) -> list[dict[str, str]]:
        """
        Separates the files to only image files.
        """
        # Create list
        files = []

        # For each file
        for file in Files:
            # Check if it's an image
            if (file["type"] == "image"):
                # Append it to the list
                files.append(file)
        
        # Return the files
        return files
    
    @staticmethod
    def GetImageInfo(Image: str | bytes) -> dict[str, any]:
        """
        Gets information of an image.
        """
        # Create image buffer
        if (isinstance(Image, str)):
            buffer = BytesIO(base64.b64decode(Image))
        elif (isinstance(Image, bytes)):
            buffer = BytesIO(Image)
        else:
            raise ValueError("Invalid image type.")
        
        # Convert image to Pillow
        image = PILImage.open(buffer)

        # Extract properties
        metdata = {
            "width": image.width,
            "height": image.height
        }

        # Close the buffers
        image.close()
        buffer.close()

        # Return the metadata
        return metdata

    @staticmethod
    async def __execute_simulated_vision__(Template: str, Files: list[dict[str, str]], VisionService: Service, Index: int, ForceNoConnect: bool) -> str:
        """
        Executes the simulated vision service and gets a response.
        """
        # Get only the images files
        files = SimulatedVisionV1.SeparateFiles(Files)

        # Check files
        if (len(files) == 0):
            return ""
        
        # For each file
        img = 1
        response = ""

        for file in files:
            # Send the message
            res = ExecuteService("", [file], VisionService, Index, ForceNoConnect, False)

            # For each token
            async for token in res:
                # Check if the response ended
                if (token["ended"]):
                    # Break the loop
                    break

                # Get the response from the service and apply to the template
                response += Template.replace("[IMAGE_ID]", str(img)).replace("[RESPONSE]", str(token["response"])) + "\n"
                img += 1
        
        # Strip the response
        response = response.strip()
        
        # Return the response
        return response
    
    @staticmethod
    async def ExecuteImageToText(Files: list[dict[str, str]], ForceNoConnect: bool) -> str:
        """
        Executes the ImageToText simulated vision.
        """
        # Return the response
        response = await SimulatedVisionV1.__execute_simulated_vision__(
            "Image [IMAGE_ID] description:\n```plaintext\n[RESPONSE]\n```\n",
            Files,
            Service.ImageToText,
            Conf.SimulatedVision_v1_Image2Text_Index,
            ForceNoConnect
        )
        return response
    
    @staticmethod
    async def ExecuteObjectDetection(Files: list[dict[str, str]], ForceNoConnect: bool) -> str:
        """
        Executes the ObjectDetection simulated vision.
        """
        # Return the response
        response = await SimulatedVisionV1.__execute_simulated_vision__(
            "Image [IMAGE_ID] list of objects detected and their position:\n```json\n[RESPONSE]\n```\n",
            Files,
            Service.ObjectDetection,
            Conf.SimulatedVision_v1_ObjectDetection_Index,
            ForceNoConnect
        )
        return response

class SimulatedVisionV2():
    @staticmethod
    def SeparateFiles(Files: list[dict[str, str]]) -> list[dict[str, str]]:
        """
        Separates the files to only video files.
        """
        # Create list
        files = []

        # For each file
        for file in Files:
            # Check if it's a video
            if (file["type"] == "video"):
                # Append it to the list
                files.append(file)
        
        # Return the files
        return files
    
    @staticmethod
    def GetVideoInfo(VideoData: bytes | str) -> dict[str, any]:
        """
        Gets information of a video.
        """
        # Check if the video data is valid
        if (isinstance(VideoData, str)):
            VideoData = base64.b64decode(VideoData)
        elif (not isinstance(VideoData, bytes)):
            raise ValueError("VideoData must be bytes or str.")
        
        # Create buffer and variables
        videoBuffer = BytesIO(VideoData)

        try:
            # Open video metadata
            reader = av.open(videoBuffer)
            stream = next(s for s in reader.streams if (s.type == "video"))

            # Extract properties
            fps = float(stream.average_rate) if (stream.average_rate) else 0
            size = [stream.codec_context.width, stream.codec_context.height]
            numberOfFrames = stream.frames
            duration = float(reader.duration / av.time_base) if (reader.duration) else (numberOfFrames / fps if (fps) else 0)

            # Save into the dictionary
            metdata = {
                "fps": fps,
                "width": size[0],
                "height": size[1],
                "frame_count": numberOfFrames,
                "duration": duration
            }

            # Close the buffers
            reader.close()
            videoBuffer.close()

            # Return the metadata
            return metdata
        except Exception as ex:
            # Close the buffer
            videoBuffer.close()

            # Raise exception
            raise Exception("Unable to get video info.") from ex

    @staticmethod
    def GetVideoFrames(VideoData: bytes | str, GetFrameEvery: int = 1) -> list[str]:
        """
        Gets the frames of the video.
        """
        # Check if the video data is valid
        if (isinstance(VideoData, str)):
            VideoData = base64.b64decode(VideoData)
        elif (not isinstance(VideoData, bytes)):
            raise ValueError("VideoData must be bytes or str.")
        
        # Create buffer
        videoBuffer = BytesIO(VideoData)

        try:
            # Open video
            reader = av.open(videoBuffer)
            frames = []
            frameCount = 0

            # For each frame
            for frame in reader.decode(video = 0):
                # Check if the frame should be extracted
                if (frameCount % GetFrameEvery == 0):
                    # Encode to JPG buffer
                    buffer = frame.to_ndarray(format = "bgr24")
                    success, buffer = cv2.imencode(".jpg", buffer)

                    # Check there are no errors
                    if (not success):
                        continue

                    # Encode the buffer and save it
                    buffer = base64.b64encode(buffer).decode("utf-8")
                    frames.append(buffer)
                
                # Increment frame count
                frameCount += 1
            
            # Close the buffers
            reader.close()
            videoBuffer.close()

            # Return the frames
            return frames
        except Exception as ex:
            # Close the buffer
            videoBuffer.close()

            # Raise exception
            raise Exception("Unable to open video.") from ex
    
    @staticmethod
    async def ExecuteSimulatedVision(Files: list[dict[str, str]], ForceNoConnect: bool) -> str:
        """
        Executes the simulated vision.
        """
        # Create variables
        simulatedVision = ""

        # Get only the video files
        files = SimulatedVisionV2.SeparateFiles(Files)

        # Check files
        if (len(files) == 0):
            return ""

        # For each video
        for video in files:
            # Get the video bytes
            videoData = video["data"]

            # Get the frames of the video
            frames = SimulatedVisionV2.GetVideoFrames(videoData, Conf.SimulatedVision_v2_Video_ProcessFrames)
            
            # Create result string variable
            simulatedVisionResult = ""

            # For each frame
            for frameIdx, frame in enumerate(frames):
                # Use img2text from simulated vision 1
                if (Conf.SimulatedVision_v1_Image2Text_Allow):
                    try:
                        # Execute the service
                        res = ExecuteService("", [{"type": "image", "data": frame}], Service.ImageToText, Conf.SimulatedVision_v1_Image2Text_Index, ForceNoConnect, False)
                        strRes = ""

                        # Receive all tokens and save into the string
                        async for token in res:
                            if (len(token["errors"]) > 0):
                                continue
                            
                            strRes += token["response"]
                        
                        # Save into the result variable
                        if (len(strRes.strip()) > 0):
                            simulatedVisionResult += f"Frame {frameIdx * Conf.SimulatedVision_v2_Video_ProcessFrames} description: {strRes}\n"
                    except:
                        pass
            
                if (Conf.SimulatedVision_v1_ObjectDetection_Allow):
                    try:
                        # Execute the service
                        res = ExecuteService("", [{"type": "image", "data": frame}], Service.ObjectDetection, Conf.SimulatedVision_v1_ObjectDetection_Index, ForceNoConnect, False)
                        strRes = ""

                        # Receive all tokens and save into the string
                        async for token in res:
                            if (len(token["errors"]) > 0):
                                continue
                            
                            strRes += token["response"]
                        
                        # Save into the result variable
                        if (len(strRes.strip()) > 0):
                            simulatedVisionResult += f"Frame {frameIdx * Conf.SimulatedVision_v2_Video_ProcessFrames} objects detected: {strRes}\n"
                    except:
                        pass
            
            # Append the simulated vision of the video
            simulatedVision += f"Video {files.index(video) + 1} frames:\n```plaintext\n{simulatedVisionResult.strip()}\n```\n"
        
        return simulatedVision
    
    @staticmethod
    async def __execute_simulated_vision_with_audition(Files: dict[str, str], ForceNoConnect: bool) -> str:
        pass  # TODO

class SimulatedAuditionV1():
    @staticmethod
    def SeparateFiles(Files: list[dict[str, str]]) -> list[dict[str, str]]:
        """
        Separates the files to only audio files.
        """
        # Create list
        files = []

        # For each file
        for file in Files:
            # Check if it's an audio
            if (file["type"] == "audio"):
                # Append it to the list
                files.append(file)
        
        # Return the files
        return files

    @staticmethod
    async def __execute_simulated_audition__(Template: str, Files: list[dict[str, str]], AuditionService: Service, Index: int, ForceNoConnect: bool) -> str:
        """
        Executes the simulated audition service and gets a response.
        """
        # Get only the audio files
        files = SimulatedAuditionV1.SeparateFiles(Files)

        # Check files
        if (len(files) == 0):
            return ""
        
        # For each file
        aud = 1
        response = ""

        for file in files:
            # Send the message
            res = ExecuteService("", [file], AuditionService, Index, ForceNoConnect, False)

            # For each token
            async for token in res:
                # Check if the response ended
                if (token["ended"]):
                    # Break the loop
                    break

                # Get the response from the service and apply to the template
                response += Template.replace("[AUDIO_ID]", str(aud)).replace("[RESPONSE]", str(token["response"])) + "\n"
                aud += 1
        
        # Strip the response
        response = response.strip()
        
        # Return the response
        return response
    
    @staticmethod
    async def ExecuteSpeechToText(Files: list[dict[str, str]], ForceNoConnect: bool) -> str:
        """
        Executes the SpeechToText simulated audition.
        """
        # Return the response
        response = await SimulatedAuditionV1.__execute_simulated_audition__(
            "> Audio [AUDIO_ID] dialogue: [RESPONSE]\n",
            Files,
            Service.SpeechToText,
            Conf.SimulatedAudition_v1_SpeechToText_Index,
            ForceNoConnect
        )
        return response

def __update_config__() -> None:
    """
    Updates the configuration in the server connection.
    """
    ServerCon.Conf = ServerCon.Conf.__from_dict__(Conf.__to_dict__())

async def ExecuteCommand(
        Service: str,
        Prompt: str = "",
        Index: int | None = None
    ) -> AsyncIterator[dict[str, any]]:
    """
    Executes a very-basic command on the server.
    You can use this, for example, to delete your conversation or get the queue for a service.
    """
    # Update config
    __update_config__()

    # Set index
    if (Index is None):
        Index = Conf.DefaultIndex

    # Serialize a very basic, minimum, data to send to the server
    jsonData = json.dumps({
        "APIKey": Conf.ServerAPIKey,
        "Prompt": Prompt,
        "Files": [],
        "Service": Service,
        "Conversation": Conf.Chatbot_Conversation,
        "Index": Index,
        "PublicKey": ServerCon.PublicKey.decode("utf-8"),
        "MaxLength": Conf.MaxLength,
        "Temperature": Conf.Temperature,
        "AllowDataSave": Conf.AllowDataSave,
        "SeeTools": Conf.SeeTools,
        "TopP": Conf.Chatbot_TopP,
        "TopK": Conf.Chatbot_TopK,
        "MinP": Conf.Chatbot_MinP,
        "TypicalP": Conf.Chatbot_TypicalP
    })

    # Return the response (probably serialized)
    async for token in ServerCon.SendAndWaitForStreaming(jsonData):
        yield token

async def ExecuteService(
        Prompt: str,
        Files: list[dict[str, str]],
        ServerService: Service,
        Index: int | None = None,
        ForceNoConnect: bool = False,
        FilesPath: bool = True
    ) -> AsyncIterator[dict[str, any]]:
    """
    This will automatically get a response a server.
    If the service requires it, this will serialize your prompt.
    This will also do other things before and after I4.0's response to your prompt.

    If FilesPath is False, the Files list must have the bytes of the files, in base64.
    If ForceNoConnect is True, this will not connect to any server, so you must be connected to one first.
    """
    # Update config
    __update_config__()

    # Set index
    if (Index is None):
        Index = Conf.DefaultIndex

    # Connect to a server
    currentServer = ServerCon.Connected[1] if (ServerCon.Connected is not None) else None

    if (not ForceNoConnect):
        if (ServerCon.IsConnected()):
            # Get the services from the server
            serverTasks = await ServerCon.GetServicesFromServer()

            if (ServerService not in serverTasks):
                # Connect to the first server to the service
                server = await ServerCon.FindFirstServer(ServerService)
                await ServerCon.Connect(server)
        else:
            # Connect to the first server to the service
            server = await ServerCon.FindFirstServer(ServerService)
            await ServerCon.Connect(server)
    elif (not ServerCon.IsConnected()):
        raise Exception("Please connect to a server first or set `ForceNoConnect` to false.")

    # Copy the files list to a new one
    files = []
    systemPrompt = Conf.Chatbot_ExtraSystemPrompts

    for file in Files:
        if (FilesPath):
            # Check if the file exists
            if (not (os.path.exists(file["data"]) and os.path.isfile(file["data"]))):
                # Return an error
                raise Exception("File doesn't exists!")

            # Read the file and convert it to base64
            with open(file["data"], "rb") as f:
                files.append({
                    "type": file["type"],
                    "data": base64.b64encode(f.read()).decode("utf-8")
                })
        else:
            # Add the file as it is
            files.append(file)

    # Serialize the data to send to the server (in some services)
    if (ServerService == Service.ImageGeneration):
        # Template: text2img
        # Set variables
        prompt = ""
        nPrompt = ""

        if (Prompt.count(" [NEGATIVE] ") > 0):
            # Contains negative prompt, set prompt and negative prompt
            prompt = Prompt[0:Prompt.index(" [NEGATIVE] ")]
            nPrompt = Prompt[Prompt.index(" [NEGATIVE] ") + 12:]
        else:
            # Doesn't contains negative prompt, set prompt
            prompt = Prompt

        # Set prompt to the Text2Image template
        Prompt = json.dumps({
            "prompt": prompt,
            "negative_prompt": nPrompt,
            "width": Conf.Text2Image_Width,
            "height": Conf.Text2Image_Height,
            "guidance": Conf.Text2Image_GuidanceScale,
            "steps": Conf.Text2Image_Steps
        })
    elif (ServerService == Service.RVC):
        # Template: RVC
        Prompt = json.dumps({
            "filter_radius": Conf.RVC_FilterRadius,
            "f0_up_key": Conf.RVC_f0,
            "protect": Conf.RVC_Protect,
            "index_rate": Conf.RVC_IndexRate,
            "mix_rate": Conf.RVC_MixRate
        })
    elif (ServerService == ServerService.UVR):
        # Template: UVR
        Prompt = json.dumps({
            "agg": Conf.UVR_Agg
        })
    elif (ServerService == Service.Chatbot):
        # Check if the chatbot is multimodal
        chatbotMultimodal = ""
        simulatedPrompt = ""

        async for token in ExecuteCommand("is_chatbot_multimodal", "", Index):
            # Check if the response is an error
            if (token["response"].lower().startswith("error") or len(token["errors"]) > 0):
                # Error, break the loop
                break

            # Not an error, set string
            chatbotMultimodal += token["response"].lower()

        if ("image" not in chatbotMultimodal and len(SimulatedVisionV1.SeparateFiles(files)) > 0):
            # Model doesn't support images. Use Simulated Vision V1

            # Use the `img2text` service if allowed in the configuration
            if (Conf.SimulatedVision_v1_Image2Text_Allow):
                try:
                    res = await SimulatedVisionV1.ExecuteImageToText(files, ForceNoConnect)
                    
                    if (len(res) > 0):
                        simulatedPrompt += res
                except:
                    pass  # Error. Ignore
            
            # Use the `od` service if allowed in the configuration
            if (Conf.SimulatedVision_v1_ObjectDetection_Allow):
                try:
                    res = await SimulatedVisionV1.ExecuteObjectDetection(files, ForceNoConnect)
                    
                    if (len(res) > 0):
                        simulatedPrompt += res
                except:
                    pass  # Error. Ignore
            
            # Separate the images
            imgs = SimulatedVisionV1.SeparateFiles(files)

            # Get the metadata of each image and save into the simulated prompt
            for imgIdx, img in enumerate(imgs):
                imgMetadata = SimulatedVisionV1.GetImageInfo(img["data"])
                simulatedPrompt += f"Image {imgIdx + 1} metadata:\n```json\n{json.dumps(imgMetadata, indent = 4)}\n```\n"

            # Remove all the images from the files
            for file in files:
                if (file["type"] == "image"):
                    files.remove(file)
        
        if ("video" not in chatbotMultimodal and len(SimulatedVisionV2.SeparateFiles(files)) > 0):
            # Model doesn't support videos. Use Simulated Vision V2

            if (Conf.SimulatedVision_v2_Video_Allow):
                try:
                    res = await SimulatedVisionV2.ExecuteSimulatedVision(files, ForceNoConnect)
                    
                    if (len(res) > 0):
                        simulatedPrompt += res
                except Exception as ex:
                    pass  # Error. Ignore
            
            # Separate the videos
            vids = SimulatedVisionV2.SeparateFiles(files)

            # Get the metadata of each video and save into the simulated prompt
            for vidIdx, vid in enumerate(vids):
                vidMetadata = SimulatedVisionV2.GetVideoInfo(vid["data"])
                simulatedPrompt += f"Video {vidIdx + 1} metadata:\n```json\n{json.dumps(vidMetadata, indent = 4)}\n```\n"
            
            # Remove all the videos from the files
            for file in files:
                if (file["type"] == "video"):
                    files.remove(file)
        
        if ("audio" not in chatbotMultimodal and len(SimulatedAuditionV1.SeparateFiles(files)) > 0):
            # Model doesn't support audios. Use Simulated Audition V1

            # Use the `speech2text` service if allowed in the configuration
            if (Conf.SimulatedAudition_v1_SpeechToText_Allow):
                try:
                    res = await SimulatedAuditionV1.ExecuteSpeechToText(files, ForceNoConnect)
                    
                    if (len(res) > 0):
                        simulatedPrompt += res
                except:
                    pass  # Error. Ignore
            
            # Remove all the audios from the files
            for file in files:
                if (file["type"] == "audio"):
                    files.remove(file)

        # Add into the system prompts
        if (len(simulatedPrompt.strip()) > 0):
            systemPrompt += f"\n{simulatedPrompt.strip()}"
        
        # Reconnect to the server if disconnected
        if (ServerCon.Connected is not None and currentServer is not None and ServerCon.Connected[1] != currentServer and not ForceNoConnect):
            await ServerCon.Connect(currentServer)

    # Set prompt
    Prompt = json.dumps({
        "Service": ServiceManager.ToString(ServerService),
        "Prompt": Prompt,
        "Files": files,
        "APIKey": Conf.ServerAPIKey,
        "Conversation": Conf.Chatbot_Conversation,
        "AIArgs": Conf.Chatbot_AIArgs,
        "SystemPrompts": systemPrompt.strip(),
        "UseDefaultSystemPrompts": Conf.Chatbot_AllowServerSystemPrompts,
        "Index": Index,
        "PublicKey": ServerCon.PublicKey.decode("utf-8"),
        "AllowedTools": Conf.AllowedTools,
        "ExtraTools": Conf.ExtraTools,
        "MaxLength": Conf.MaxLength,
        "Temperature": Conf.Temperature,
        "SeeTools": Conf.SeeTools,
        "AllowDataSave": Conf.AllowDataSave,
        "TopP": Conf.Chatbot_TopP,
        "TopK": Conf.Chatbot_TopK,
        "MinP": Conf.Chatbot_MinP,
        "TypicalP": Conf.Chatbot_TypicalP
    })

    # Return the response
    async for token in ServerCon.SendAndWaitForStreaming(Prompt):
        yield token

async def GetQueueForService(QueueService: Service, Index: int = -1) -> tuple[int, int]:
    """
    This will send a prompt to the connected server asking for the queue size and time.
    Once received, it will return it.

    Index == -1 automatically gets the model with the smallest queue size.
    """

    # Get response from the queue command
    res = ExecuteCommand("get_queue", ServiceManager.ToString(QueueService), Index)

    async for response in res:
        if (not response["ended"]):
            # Continue, ignoring this message
            continue

        # Deserialize the received JSON response to a dictionary and return the users and the time
        queue = json.loads(response["response"])
        return (queue["users"], queue["time"])

    # Throw an error
    raise Exception("Queue error: Error getting queue.")

async def DeleteConversation(Conversation: str | None) -> None:
    """
    This will delete your current conversation ONLY on the connected server.
    If `Conversation` is null this will use the conversation of the configuration.
    """

    CConversation = Conf.Chatbot_Conversation

    if (Conversation != None):
        # Update the conversation in the configuration settings
        Conf.Chatbot_Conversation = Conversation

    # Send request to the server and get the result
    res = ExecuteCommand("clear_conversation")

    # Restore the conversation in the configuration settings
    Conf.Chatbot_Conversation = CConversation

    async for response in res:
        if (not response["ended"]):
            # Continue, ignoring this message
            continue

        # Set response variable
        result = response["response"].lower().strip()

        # Check if response it's invalid
        if (result != "conversation deleted." or len(response["errors"]) > 0):
            # It's invalid, throw an error
            raise Exception(f"Error deleting the conversation. Got `{result}`; `conversation deleted.` expected. Errors: {response['errors']}")

        # It's a valid response, return
        return

    # Throw an error
    raise Exception("Delete conversation error.")

async def DeleteMemory(Memory: int = -1) -> None:
    """
    This will delete your current memory/memories ONLY on the connected server.
    If `Memory` is -1 this will delete all the memories.
    """

    if (Memory == -1):
        # Set the command to delete all the memories
        cmd = "clear_memories"
    else:
        # Set the command to delete a memory
        cmd = "clear_memory"

    # Send request to the server and get the result
    res = ExecuteCommand(cmd, str(Memory))

    async for response in res:
        if (not response["ended"]):
            # Continue, ignoring this message
            continue

        # Set response variable
        result = response["response"].lower().strip()

        # Check if response it's invalid
        if ((result != "memories deleted." and result != "memory deleted.") or len(response["errors"]) > 0):
            # It's invalid, throw an error
            raise Exception(f"Error deleting the memories/memory. Got `{result}`; `memories deleted.` or `memory deleted.` expected. Errors: {response['errors']}")

        # It's a valid response, return
        return

    # Throw an error
    raise Exception("Delete memory/memories error.")

async def GetTOS() -> str:
    """
    Gets the server's Terms Of Service.
    """

    # Send command to the server and wait for response
    res = ExecuteCommand("get_tos")

    # For each response
    async for response in res:
        # Check if it ended
        if (not response["ended"]):
            # Continue, ignoring this message
            continue

        # It ended, get the text response
        tResponse = response["response"]

        # Strip the response
        tResponse = tResponse.strip()

        # Check the length of the response
        if (len(tResponse) == 0):
            # There are not TOS
            return "No TOS."

        # There are TOS, return the text response
        return tResponse

    raise Exception("Error getting TOS: No response from server.")